/** File CacheChain.java
* Classe protetta Java che implementa 
* la catena degli elementi della cache.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

class CacheChain implements ICacheChain
{
   /** classe privata CacheChainNode
   *   Si tratta di un nodo della catena
   */
   private class CacheChainNode
   {
      public int status = CacheConstants.I;
      public CacheChainNode left = null;
      public CacheChainNode right = null;
      public ICacheElem trdata = null;

   }   // CacheChainNode
   
   // membri
   private CacheChainNode validnode = null;
   private CacheChainNode invalidnode = null;
   private int size = 0;
   
   // costruttori
   public CacheChain()
   {
      this(CacheConstants.DEFAULT_CACHE_SIZE);
   }
   
   public CacheChain( int size )
   {
      this.size = size;
      if ( size <= 0 ) throw new NullCacheSizeException();
      size--;
      CacheChainNode node = new CacheChainNode();
      validnode = node;
      invalidnode = node;
      while( size > 0 )
      {
         size--;
         node = new CacheChainNode();
         
         // inserisco il nodo nei nodi invalidi
         // della catena
         validnode.right = node;
         node.left = validnode;
         validnode = node;
      }
   }
   
   // ritorna !=null in caso di hit modified
   public synchronized ICacheElem shareData( ITag tag )
   {
      // scandiamo la catena a partire dai nodi validi
      CacheChainNode node = validnode;
      while((node!=null)&&(node.status!=CacheConstants.I))
         if (node.trdata.getTag().isEqual(tag))
         {
            // abbiamo trovato il nodo da rendere condiviso
            // se esso era modified dobbiamo ritornarlo
            int oldstatus = node.status;
            node.status = CacheConstants.S;
            if( oldstatus == CacheConstants.M ) return node.trdata;
            else return null;
         }
         // se non  quello che cerchiamo andiamo avanti
         else node = node.left;
      
      // se siamo qui non abbiamo trovato il nodo
      return null;
   
   } // shareData()
   
   
   // ritorna !=null in caso di hit modified
   public synchronized ICacheElem invalidData(ITag tag)
   {
      // scandiamo la catena a partire dai nodi validi
      CacheChainNode node = validnode;
      while((node!=null)&&(node.status!=CacheConstants.I))
      {
         if (node.trdata.getTag().isEqual(tag))
         {
            // sposta il nodo trovato in fondo alla catena
            if(node.left!=null) // se non  gi in fondo
            {
               if(node.right!=null) // e se non  il primo
               {
                  // allora lo si rimuove dalla catena
                  node.left.right = node.right;
                  node.right.left = node.left;
               }
               else // mentre se  il primo si cancella un estremo
               {
                  node.left.right = null;
                  validnode = node.left;
               }
               
               // ora che il nodo  stato rimosso dalla catena
               // lo inseriamo in fondo
               invalidnode.left = node;
               node.right = invalidnode;
               node.left = null;
               invalidnode = node;
            }
            
            // invalida il nodo trovato 
            // tenendo traccia del suo vecchio stato
            int oldstatus = node.status;
            node.status = CacheConstants.I;
            
            // se il vecchio stato era M allora devo ritornare il dato
            if( oldstatus == CacheConstants.M )
               return node.trdata;
            else return null;
         }
         
         // se non  il nodo cercato, vai al prossimo
         node = node.left;
      } // fine ciclo
      
      // se siamo qui non abbiamo trovato nulla
      return null;
   
   } // invalidData()
   
   
   // ritorna null in caso di miss
   public synchronized ICacheElem getData( ITag tag )
   {
      // scandiamo la catena a partire dai nodi validi
      CacheChainNode node = validnode;
      while((node!=null)&&(node.status!=CacheConstants.I))
      {
         if (node.trdata.getTag().isEqual(tag))
         {
            // sposta il nodo trovato in cima alla lista
            if(node.right!=null) // se non  gi in cima
            {
               if(node.left!=null) // e se non  l'ultimo
               {
                  // allora lo si sposta
                  node.left.right = node.right;
                  node.right.left = node.left;
               }
               else // mentre se  l'ultimo si cancella un estremo
               {
                  node.right.left = null;
                  invalidnode = node.right;
               }
               
               // ora che il nodo  stato rimosso dalla catena
               // lo inseriamo in cima
               validnode.right = node;
               node.left = validnode;
               node.right = null;
               validnode = node;
            }
            
            // ritorna il nodo trovato
            return node.trdata;
         }
         
         // se non  il nodo cercato, vai al prossimo
         node = node.left;
      } // fine ciclo
      
      // se siamo qui non abbiamo trovato nulla
      return null;
   
   } // getData()
   
      
   // ritorna !=null in caso di hit modified
   public synchronized ICacheElem insertData( ICacheElem trdata, boolean shared )
   {
      // estraiamo dalla catena l' ultimo nodo
      // esso  quello pi probabilmente invalido
      CacheChainNode node = invalidnode;
      if( invalidnode.right != null ) 
      {
         invalidnode.right.left = null;
         invalidnode = invalidnode.right;
         node.right = null;
      }
      
      // memorizziamone i vecchi dati
      int oldstatus = node.status;
      ICacheElem oldlink = node.trdata;
      
      // inseriamo i nuovi dati
      node.trdata = trdata;
      node.status = ( shared )? CacheConstants.S : CacheConstants.E ;
      
      // inseriamo il nodo in cima alla lista
      if( validnode != node )
      {
         validnode.right = node;
         node.left = validnode;
         validnode = node;
      }
      
      // se il vecchio stato era M dobbiamo ritornarlo
      
      return ( oldstatus == CacheConstants.M )? oldlink : null ;
   
   } // insertData()
   
   
   public synchronized ICacheElem[] flushData()
   {
      ICacheElem[] temp = new ICacheElem[getSize()];
      int index = 0;
      // scandiamo la catena a partire dai nodi validi
      CacheChainNode node = validnode;
      while((node!=null)&&(node.status!=CacheConstants.I))
      {
         // se il nodo  modified lo memorizziamo
         if(node.status == CacheConstants.M)
         {
            temp[index] = node.trdata;
            index++;
         }
         // azzeriamo la catena
         node.status = CacheConstants.I;
         node.trdata = null;
         // avanti il prossimo
         node = node.left;
      }  
      // ora index indica quanti elementi modified abbiamo raccolto
      // in caso di raccolto magro ritiriamoci con la
      // coda tra le gambe
      if( index == 0 ) return null;
      ICacheElem[] res = new ICacheElem[index];
      // li copiamo dove ci pare
      for( int j = 0; j < index; j++ )
         res[j] = temp[j];
      return res;
   
   }  // flushData()  
   
   
   public int getSize() { return size; }
   
   // ritorna lo stato di un dato, I se miss
   public synchronized int getStatus( ITag tag )
   {
      // scandiamo la catena a partire dai nodi validi
      CacheChainNode node = validnode;
      while((node!=null)&&(node.status!=CacheConstants.I))
      {
         if (node.trdata.getTag().isEqual(tag))
         {
            // sposta il nodo trovato in cima alla lista
            if(node.right!=null) // se non  gi in cima
            {
               if(node.left!=null) // e se non  l'ultimo
               {
                  // allora lo si sposta
                  node.left.right = node.right;
                  node.right.left = node.left;
               }
               else // mentre se  l'ultimo si cancella un estremo
               {
                  node.right.left = null;
                  invalidnode = node.right;
               }
               
               // ora che il nodo  stato rimosso dalla catena
               // lo inseriamo in cima
               validnode.right = node;
               node.left = validnode;
               node.right = null;
               validnode = node;
            }
            
            // ritorna lo stato del nodo trovato
            return node.status;
         }
         
         // se non  il nodo cercato, vai al prossimo
         node = node.left;
      } // fine ciclo
      
      // se siamo qui non abbiamo trovato nulla
      return CacheConstants.I;
   
   } // getStatus()
   
   
   // setta lo stato di un dato e ritorna il precedente, I se miss
   public synchronized int setStatus( ITag tag, int newstat )
   {
      int oldstat = getStatus(tag);
      if( oldstat == CacheConstants.I ) return oldstat ;
      // se c' un hit invece ...
      // scandiamo la catena a partire dai nodi validi
      // finch non trovi quello buono
      CacheChainNode node = validnode;
      while((node!=null)&&(node.trdata.getTag().isEqual(tag)))
         node = node.left;
      // ed a quel punto colpisci
      node.status = newstat;
      return oldstat;
   
   } // setStatus()
 
   
   // stampa la catena
   public synchronized String toString()
   {
      // scandiamo la catena a partire dai nodi validi
      CacheChainNode node = validnode;
      String str = "Chain holds:\n";
      while((node!=null)&&(node.status!=CacheConstants.I))
      {
         str = str + node.trdata.getTag().toString() + "\n";
         node = node.left;
      }
      return str;
   }
   
}   // CacheChain
