/** File CacheServer.java
* Classe astratta Java che implementa  
* la main architecture del server di una cache.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;
import java.net.*;

public abstract class CacheServer implements IServer,
                                             Runnable,
                                             IPersistentData,
                                             ICacheDataLoader
{
   // membri
   // lista dei dati attualmente in uso 
   private ILispList datainuse = LispUtils.EmptyList();
   // lista dei dati attualmente in group actions
   private ILispList datainsnoop = LispUtils.EmptyList();
   // lock di accesso alle liste
   private Boolean Llock = new Boolean(true);
   // lista dei clienti connessi
   private ILispList clientlist = LispUtils.EmptyList();
   // lock della lista
   private Boolean CLock = new Boolean(true);
   // se il server si deve fermare
   private boolean servermuststop = false;
   // nome del server
   private String servername = null;
   // indica se il server  il master
   private boolean masterserver = false;
   // in caso di terminazione del server indica che  stata
   // causata da un ordine esterno
   protected boolean endcausedfrommasterserver = false;
   // porta del server daemon
   private int serverport = 0;
   // lista dei server fratelli:  ordinata secondo 
   // la priorit degli altri server.
   // il primo della lista  il master a meno che non sia
   // masterserver == true
   private ILispList brothersInfo = null;
   // lista dei fratelli riconosciuti: ogni slave che si
   // connette deve apparire qui
   private ILispList expectedbros = null;
   // ID del server 
   private long serverID;
   // priorit del server: pu servire per gestire alcune azioni
   // di gruppo
   private int priority;
   // socket di ascolto
   private ServerSocket listener = null;
   // unit per la consistenza dei dati
   private CacheCLMDNR clogic = null;
   // costruttore
   public CacheServer( String servername, 
                       long serverID,
                       int serverport,
                       boolean masterserver,
                       ILispList brothersInfo,
                       int priority ) throws IOException
   {
      this.servername = servername;
      this.serverport = serverport;
      this.masterserver = masterserver;
      this.expectedbros = brothersInfo;
      this.brothersInfo = LispUtils.EmptyList();
      this.serverID = serverID;
      this.priority = priority;
      
      listener = new ServerSocket(serverport);
      
      // procedura per lo starting del server
      if(masterserver) MasterStart();
      else SlaveStart();
      
      // ora che siamo partiti attendiamo connessioni da clienti
      // pertanto impostiamo il necessario timeout
      listener.setSoTimeout(CacheConstants.ACCEPT_POLLING_TIME);
      
      // infine creiamo una unit di consistenza 
      // dei dati
      clogic = createConsistencyController();
   }
   
   private void MasterStart() throws IOException
   {
      System.out.println("Master Server Starting.\nAllowed slaves are:\n"+expectedbros.toSimpleString());
      if(expectedbros.isNull()) 
         System.out.println("none");
      
   } // MasterStart()
     
   private void SlaveStart() throws IOException
   {
      System.out.println("Slave server starting...");
      // chi  il master ?
      brothersInfo = expectedbros;
      CacheServerInfo masterinfo = 
               ((CacheServerInfo)(brothersInfo.car()));
      System.out.println("Master server is :");
      System.out.println(masterinfo.toSimpleString());
      // connettiamoci ad esso
      System.out.println("Connecting...");
      Socket conn = new Socket(masterinfo.getAddress(),
                               masterinfo.getPort());
      DataInputStream datain = 
                  new DataInputStream(conn.getInputStream());
      DataOutputStream dataout = 
                  new DataOutputStream(conn.getOutputStream());
      // inviamogli una richiesta di connessione tra servers
      System.out.println("Sending informations...");
      dataout.writeShort(CacheConstants.SLAVE_IS_CALLING);
      System.out.println("Sending ID...");
      dataout.writeLong(serverID);
      dataout.flush();
      // attendiamo la risposta
      System.out.println("Sent. Waiting for reply...");
      datain.readShort(); // se siamo qui va tutto bene
      System.out.println("Master accepted us.");
   
   } // SlaveStart()
   
   // indica che il thread server si deve fermare
   public synchronized boolean ServerMustStop()
   { return servermuststop; }
   
   // ferma il server
   public synchronized void StopServerThreads( boolean causeIsAnotherServer )
   { 
      servermuststop = true;
      endcausedfrommasterserver = causeIsAnotherServer;
   }
   
   public void waitUntilThereAreNoMoreClients(Thread launcher)
   {
      while(!getClientsList().isNull())
         try
         {
            //debug
            System.err.println("Still waiting for clients");
            launcher.sleep(CacheConstants.SLEEPY_TIME);
         }
         catch(InterruptedException er){};
         
   }
   
   public void waitUntilThereAreNoMoreServers(Thread launcher)
   {
      while(!getBros().isNull())
         try
         {
            //debug
            System.err.println("Still waiting for servers");
            launcher.sleep(CacheConstants.SLEEPY_TIME);
         }
         catch(InterruptedException er){};
         
   }

   public CacheDataInfo AskDataForReading( ITag tag, boolean externalsnoop ) throws
                                                CollisionException,
                                                ServerPermissionException,
                                                IOException
   {
      //debug
      System.err.println("AskDataForRd/Wr started for "+tag);
      synchronized(Llock)
      {
         //debug
         System.err.println("Entering in locked section of AskDataForRd/Wr for "+tag);
         ILispList data = datainsnoop;
         boolean found = false;
         do{
         while(!found && !data.isNull())
         {
            found = ((CacheDataInfo)data.car()).getElem().getTag().isEqual(tag);
            data = data.Tail();
         }
         // collisione se lo si trova negli 
         //snoops durante uno snoop
         if ( found && externalsnoop )
         {
            //debug
            System.err.println("Data asked "+tag+" is in use by a snoop.");
            System.err.println("Launching Collision Exception.");
            throw new CollisionException();
         }
         // cerchiamolo nei dati tranquilli
         data = datainuse;
         while(!found && !data.isNull())
         {
            found = ((CacheDataInfo)data.car()).getElem().getTag().isEqual(tag);
            data = data.Tail();
         }
         // se l'abbiamo trovato rilasciamo il lock
         if(found) 
            try
            {
               //debug
               System.err.println("Data asked "+tag+" is already in use.");
               System.err.println("Waiting for notify.");
               Llock.wait();
            }
            catch(InterruptedException ier){};
            
         } while(found);// continua finch il dato non  pi in lista
         
         // adesso leggi il dato dallo storage
         try
         {
            CacheDataInfo dt = LoadData(tag);
            //debug
            System.err.println("Data asked "+tag+" has been found.");
            System.err.println("Inserting "+tag+" in list "+
                               (externalsnoop?" of snoops.":" of use."));
            if (externalsnoop) datainsnoop = 
                     LispUtils.cons(dt, datainsnoop);
            else datainuse = LispUtils.cons( dt, datainuse );
            return dt;
         }
         catch(DataIsMissingException e)
         {
            // verr eventualmente provocato uno snoop che 
            // dovr poi RILASCIARE il dato in questione
            if(!externalsnoop) // infatti se avessi uno snoop
                               // esso terminerebbe lanciando
                               // MISS
            {
               // costruiamo un dato fake da mettere in lista
               //debug
               System.err.println("Data asked "+tag+" is not in server.");
               System.err.println("Preparing for a snoop.");
               ICacheElem fk = new FakeData(tag);
               CacheDataInfo fake = new CacheDataInfo(fk, null, null, null);
               datainsnoop = LispUtils.cons(fake, datainsnoop);
            }
            return null;
         }
      }  
   }
                                                   
   public CacheDataInfo AskDataForWriting( ITag tag, boolean externalsnoop ) throws
                                                CollisionException,
                                                ServerPermissionException,
                                                IOException
   {
      // poich anche qui le condizioni sono le stesse
      return AskDataForReading(tag, externalsnoop);
   }
   
   // NOTA : NON SCRIVE SU DISCO NIENTE!!!!
   public void ReleaseData( ITag tag )
   {
      CacheDataInfo data = new CacheDataInfo( new FakeData(tag), null, null, null);
      // rilascia il lock sul dato
      synchronized(Llock)
      {
         //debug
         System.err.println("Removing "+tag+" from data in use/snoop.");
         datainuse = LispUtils.RemoveElem(datainuse, data);
         datainsnoop = LispUtils.RemoveElem(datainsnoop, data);
         Llock.notifyAll();
      }
   }                                            

   protected abstract CacheDataInfo LoadData( ITag tag ) throws 
                                                DataIsMissingException,
                                                ServerPermissionException,
                                                IOException;
      
   // serve a caricare l'ultimo valore valido rimasto sul server
   // a differenza di loaddata la DataIsMissingException 
   // lanciata solo se proprio il dato NON c'
   public abstract CacheDataInfo LoadLastValidValue( ITag tag ) throws 
                                                DataIsMissingException,
                                                ServerPermissionException,
                                                IOException;
   public abstract void SaveDataToStorage(CacheDataInfo dat) throws IOException;
   
   protected abstract void MarkDataAsInvalid(ITag tag, CacheServerInfo newsource)
                                                throws IOException;
                                                   
   protected abstract void MarkDataAsInvalidFromCache(ITag tag, CacheClientInfo newsource)
                                                throws IOException;
                                                
   protected abstract void RestoreMarkedAsInvalidDatas() throws IOException;
   
   protected abstract CacheCLMDNR createConsistencyController();
   
   public void run()
   {
      Socket sock;
      try
      {
         while(!ServerMustStop())
         {
            sock = null;
            try
            {
               sock = listener.accept();
            }
            catch( InterruptedIOException ie )
            {
               if (ServerMustStop()) 
               {
                  System.out.println("Closing and terminating server...");
                  TerminatingServer();
                  return;
               }
            }
            
            if(sock!=null)
            {
               // se siamo qui possiamo processare la richiesta
               System.out.println("Detected connection from "+
                                  sock.getInetAddress().toString()+
                                  ":"+sock.getPort());
               Thread mythread = new ThreadCaster(sock, this);
               mythread.start();
            }
         }
         // fine ciclo
         System.out.println("Closing and terminating server...");
         TerminatingServer();
         return;
      }
      catch( IOException e )
      {
         System.err.println("Exception raised on accept():\n"+e);
      }
   
   } // run()
   
   protected void TerminatingServer()
   {
      // devo anzitutto attendere che le varie caches
      // abbiano terminato i loro snoop finali
      // questo garantisce che:
      // i dati delle caches sono tutti flushati
      // e che i dati presenti sul server locale,
      // nel caso in cui sia uno slave, corrispondano con quelli del 
      // master.
      // Ne segue che mai e poi mai il master chieder a questo slave ulteriori
      // dati. Tuttavia il master potrebbe chiedere allo slave di invalidare
      // alcuni suoi dati.
      // pertanto  necessario che lo slave informi il master
      // che si sta disconnettendo. Lo slave non  dunque tenuto
      // a dare al master ulteriori informazioni
      // nel caso fosse il master a terminare allora  necessario
      // che a seguito della richiesta di disconnessione lo slave gli invii 
      // tutti i dati invalidi presenti sulle sue caches
      
      // per prima cosa devo impedire che gli altri server
      // lancino ulteriori snoop cycles( se sono il master )
      // attendo che i client abbiano finito
      // ora devo differenziare idue casi : master o slave
      if(getMaster()) TerminateMaster();
      else TerminateSlave();
   }
   
   protected void TerminateSlave()
   {
   	// attiva il thread che attende la chiusura di ogni 
	   // cliente e poi lancia la richiesta di chiusura
      CacheServerInfo masterinfo = (CacheServerInfo)(getBros().Head());
      ClientWaiterThread cwt = new ClientWaiterThread(masterinfo);
      cwt.start();
	   AcceptServerSnoopUntilHasFinished(cwt);
	   try
	   {
	      listener.setSoTimeout(CacheConstants.TIMEOUT);
		   listener.close();
         //debug
         System.err.println("Listening socket closed.");
		}
		catch(IOException e){};
   }
   
   protected void TerminateMaster()
   {
      // attiva il thread che lancia a tutti gli slave
      // il messaggio di disconnessione
      // ed attende che tutti abbiano terminato,
      // sia che essi siano clienti, sia che essi siano server slave
      ServerTerminatorThread sterm = new ServerTerminatorThread();
      sterm.start();
	   AcceptServerSnoopUntilHasFinished(sterm);
	   try
	   {
	      // chiudi la socket di ascolto
	      listener.setSoTimeout(CacheConstants.TIMEOUT);
		   listener.close();
         //debug
         System.err.println("Listening socket closed.");
		   // notare che adesso non ci sono pi
		   // snoop in corso
		   // infatti ogni server ha inviato la propria disconnessione
   		// ora il master deve annullare le eventuali invalidazioni rimaste
   		RestoreMarkedAsInvalidDatas();
		}
		catch(IOException e){};
   }
   
   private void AcceptServerSnoopUntilHasFinished(Thread sterm)
   {
      //debug
      System.err.println("From now on only server calls will be accepted.");
      while(sterm.isAlive())
      {
         try
         {
            Socket sock = listener.accept();
            // se siamo qui possiamo processare la richiesta
            System.out.println("Detected connection from "+
                               sock.getInetAddress().toString()+
                               ":"+sock.getPort());
            Thread mythread = new ThreadCaster(sock, this, true);// impedisce ad eventuali
                                                   // nuovi clienti di entrare
            mythread.start();
         }
         catch(IOException i){};
      }
   }
   
   public ILispList getBros()
   { 
      synchronized(CLock)
      {
         return brothersInfo; 
      }
   }
   
   public ILispList getClientsList()
   {
      synchronized(CLock)
      {
         return clientlist;
      }
   }
   
   public boolean getMaster() { return masterserver; }
   
   public CacheServerInfo getInfo() throws UnknownHostException
   {
      return new CacheServerInfo(servername, serverID,
                                    InetAddress.getLocalHost(),
                                    serverport, priority);
   }
   
   public void AddClient(CacheClientInfo cl)
   {
      synchronized(CLock)
      {
         clientlist = LispUtils.cons( cl, clientlist );
      }
   }
   
   public void RemoveClient(CacheClientInfo cl)
   {
      synchronized(CLock)
      {
         clientlist = LispUtils.RemoveElem( clientlist, cl );
      }
   }
   
   protected void AddServer(CacheServerInfo sr)
   {
      synchronized(CLock)
      {
         brothersInfo = 
            LispUtils.OrderHighFirst(
               LispUtils.cons( sr, brothersInfo ));
      }
   }
   
   protected void RemoveServer(CacheServerInfo sr)
   {
      synchronized(CLock)
      {
         brothersInfo = LispUtils.RemoveElem( brothersInfo, sr );
      }
   }
   
   protected void errorInSlave(CacheServerInfo slaveinfo)
   {
      // lo snoop  fallito, non fa niente
   }

   protected void errorInMaster(CacheServerInfo masterinfo)
   {
      // lo snoop  fallito, non fa niente
   }

   private void RegisterSlave( Socket s, 
                               DataInputStream datain,
                               DataOutputStream dataout )
                               throws IOException
   {
      long sentID = datain.readLong();
      System.out.println("Detected server ID = "+sentID);
      CacheServerInfo read = 
            new CacheServerInfo( null, sentID, null, 0, 0 );
      // controlla se l'elemento  gi in lista
      boolean alreadyinlist = false;
      synchronized(CLock)
      {
         alreadyinlist = 
            ( -1 != LispUtils.getElemPos( brothersInfo, read ));
      }
      // controlla se gli  permesso entrare in lista
      int index = LispUtils.getElemPos( expectedbros, read );
            
      // se l'elemento non  in lista o non  permesso fermati
      if(( index == -1 )|| alreadyinlist ) 
      {
         System.out.println("Server ID is unknown or server already accepted. Excluding request...");
         s.close();
      }
      else // l'elemento  in lista
      {
         CacheServerInfo found = 
            (CacheServerInfo)LispUtils.getElemAtPos(expectedbros, index);
         System.out.println("Server ID accepted.");
         AddServer(found);
         System.out.print("Sending ACK to slave "+found.toSimpleString()+
                          "...");
         dataout.writeShort(CacheConstants.ACK);
         dataout.flush();
         System.out.println(" Sent.");
         s.close();
      }
   }
   
   private class ThreadCaster extends Thread
   {
      private Runnable toRun = null;
      private Socket s = null;
      private CacheServer owner = null;
      boolean notclients = false;
      
      public ThreadCaster( Socket s, CacheServer owner )
      {
         super();
         this.s = s;
         this.owner = owner;
      }
      
      public ThreadCaster( Socket s, CacheServer owner, boolean notclients )
      {
         this(s, owner);
         this.notclients = notclients;
      }
      
      public void run() 
      { 
         //debug
         System.err.println("Starting ThreadCaster.");
         try{
         DataInputStream datain = 
                  new DataInputStream(s.getInputStream());
         DataOutputStream dataout = 
                  new DataOutputStream(s.getOutputStream());
         switch(datain.readShort())
         {
         case CacheConstants.CLIENT_BEGINS_SESSION:
            if(notclients) {s.close(); return;} 
            toRun = new CacheServingThread( s, owner, datain, dataout );
            break;
         case CacheConstants.SERVER_BEGINS_SESSION:
            toRun = new ServerServingThread( s, owner, datain, dataout );
            break;
         case CacheConstants.SLAVE_IS_CALLING:
            // uno slave richiede una connessione
            if(notclients) {s.close(); return;} 
            owner.RegisterSlave( s, datain, dataout );
            break;
         default : s.close(); break;
         }
         if(toRun!=null) toRun.run(); 
         }
         catch(IOException e )
         {
            System.err.println("Exception raised on accept():\n"+e);
         }
         //debug
         System.err.println("ThreadCaster has finished.");
      }
   
   } // ThreadCaster

   private class ClientWaiterThread extends Thread
   {
      // membri
      CacheServerInfo masterinfo = null;
      
      // costruttore
      public ClientWaiterThread(CacheServerInfo masterinfo)
      {
         super();
         this.masterinfo = masterinfo;
      }
      
      // esecuzione
      public void run()
      {
         boolean open = false;
         DataInputStream datain = null;
         DataOutputStream dataout = null;
         Socket sock = null;
         
         //debug
         System.err.println("Starting ClientWaiterThread.");
         // attendi la fine di tutti i clients
         waitUntilThereAreNoMoreClients(this);
         
         // richiedi al master la disconnessione
         try
         {
            sock = new Socket(masterinfo.getAddress(), 
                                     masterinfo.getPort());
            open = true;
            // settaggio opzioni socket
            sock.setTcpNoDelay(CacheConstants.TCPNODELAY);
            sock.setSoTimeout(CacheConstants.TIMEOUT);
            sock.setSoLinger(CacheConstants.LINGER, CacheConstants.LINGERTIME);
            datain = new DataInputStream( sock.getInputStream() );
            dataout = new DataOutputStream( sock.getOutputStream() );
            // ora invio l'ordine
            dataout.writeShort(CacheConstants.SERVER_BEGINS_SESSION);
            dataout.writeLong(getInfo().getID());
            dataout.writeShort(CacheConstants.SLAVE_IS_CLOSING);
            dataout.flush();
            datain.readShort();// leggi ACK;
            sock.close();
            open = false;
         }
         catch(IOException e)
         {
            if(open) try{sock.close();}catch(IOException e2){};
         }
         //debug
         System.err.println("ClientWaiterThread has finished.");
      }
      
   } // ClientWaiterThread class
   
   private class ServerTerminatorThread extends Thread
   {
      public ServerTerminatorThread(){ super(); }
      
      public void run()
      {
         boolean open = false;
         DataInputStream datain = null;
         DataOutputStream dataout = null;
         Socket sock = null;

         //debug
         System.err.println("Starting ServerTerminatorThread.");
         // lancia a tutti i server un messaggio di 
         // chiusura
         ILispList slist = getBros();
         while(!slist.isNull())
         {
            // per ogni server lancia un messaggio
            CacheServerInfo slaveinfo = (CacheServerInfo)(slist.Head());
            // prova
            
            try
            {
               sock = new Socket(slaveinfo.getAddress(), 
                                        slaveinfo.getPort());
               open = true;
               // settaggio opzioni socket
               sock.setTcpNoDelay(CacheConstants.TCPNODELAY);
               sock.setSoTimeout(CacheConstants.TIMEOUT);
               sock.setSoLinger(CacheConstants.LINGER, CacheConstants.LINGERTIME);
               datain = new DataInputStream( sock.getInputStream() );
               dataout = new DataOutputStream( sock.getOutputStream() );
               // ora invio l'ordine
               dataout.writeShort(CacheConstants.SERVER_BEGINS_SESSION);
               dataout.writeLong(getInfo().getID());
               dataout.writeShort(CacheConstants.MASTER_IS_CLOSING);
               dataout.flush();
               datain.readShort();// leggi ACK;
               sock.close();
               open = false;
            }
            catch(IOException e)
            {
               if(open) try{sock.close();}catch(IOException e2){};
            }
            slist = slist.Tail();
         }
         
         // attendi la fine dei clients
         waitUntilThereAreNoMoreClients(this);
         // e dei servers
         waitUntilThereAreNoMoreServers(this);
         //debug
         System.err.println("ServerTerminatorThread has finished.");
      }
      
   } // ServerTerminatorThread class

   // metodi di ICacheDataLoader
   public ITag readTag( DataInputStream datain ) throws IOException
   {
      return clogic.readTag(datain);
   }
   
   public IUpdateElem readUpdateElem( DataInputStream datain ) throws IOException
   {
      return clogic.readUpdateElem(datain);
   }
   
   public ICacheElem readCacheElem( DataInputStream datain ) throws IOException
   {
      return clogic.readCacheElem(datain);
   }
   
   public ICacheElem updateIsCorrect( ICacheElem oldelem, IUpdateElem news ) 
                                          throws InconsistentUpdateException
   {
      return clogic.updateIsCorrect(oldelem, news);
   }
   
} // CacheServer

