/** File CacheServingThread.java
* Classe Java che implementa  
* un thread del server che interagisce con una cache.
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;
import java.net.*;

public class CacheServingThread extends ServingThread 
{
   // chi  il client
   private CacheClientInfo clientinfo = null;
   // qual' il comando in corso
   private short command = 0;
   
   public CacheServingThread( Socket s, CacheServer o,
                              DataInputStream datain,
                              DataOutputStream dataout )
                                       throws IOException
   {
      super(s, o, datain, dataout);
   }
   
   /* dati protected ereditati
   protected DataInputStream datain;
   protected DataOutputStream dataout;
   protected Socket sock;
   protected CacheServer owner;
   */
   
   public void run()
   {
      // prova ad inizializzare la connessione con il client
      boolean initialized = InitializeClientConnection();
      if(!initialized) return;
      
      // registriamo il cliente
      owner.AddClient(clientinfo);
      System.out.println("Client "+clientinfo.getID()+" connected.");
      
      // lettura del comando
      try
      {
         boolean exit = false;
         do
         {
            sock.setSoTimeout(CacheConstants.ACCEPT_POLLING_TIME);
            boolean cont = true;
            do
            try
            {
               command=datain.readShort();
               cont = false;
            } // pu interrompersi
            catch(InterruptedIOException ioe)
            {
               if(owner.ServerMustStop()) 
               {
                  // in uscita togliamo il cliente dalla lista
                  owner.RemoveClient(clientinfo);
                  // flushiamo i dati del cliente
                  LaunchFinalSnoopFlush();
                  return;
               }
            }
            while(cont);//attendi una chiamata 
            
            exit = command == CacheConstants.CLIENT_DISCONNECTING;
            sock.setSoTimeout(CacheConstants.TIMEOUT);
            if(!exit)
            switch(command)
            {
            // comandi standard
            case CacheConstants.LINEFILL_CACHEABLE :
               System.out.println("Client "+clientinfo.getID()+" queries a cacheable linefill.");
               LineFill(true);
               break;
            case CacheConstants.LINEFILL_UNCACHEABLE :
               System.out.println("Client "+clientinfo.getID()+" queries an uncacheable linefill.");
               LineFill(false);
               break;
            case CacheConstants.WRITETHROUGH :
               System.out.println("Client "+clientinfo.getID()+" queries a writethrough.");
               WriteThrough();
               break;
            case CacheConstants.WRITE_EXCLUSIVE :
               System.out.println("Client "+clientinfo.getID()+" asks to use a data exclusively.");
               WriteExclusive();
               break;
            case CacheConstants.UPDATE :
               System.out.println("Client "+clientinfo.getID()+" asks to update a data.");
               Update();
               break;
            default : // comando sconosciuto
               System.out.println("Client "+clientinfo.getID()+" asks an unknown command ("+command+").");
               dataout.writeShort(CacheConstants.NACK);
               dataout.flush();
               throw new IOException();
            }
         }
         while(!exit);
         // gestione della disconnessione
         System.out.println("Client "+clientinfo.getID()+" disconnecting.");
         Disconnecting();
         owner.RemoveClient(clientinfo);
      }
      catch( IOException e )
      {   
         System.err.print("An error occoured during interaction");
         e.printStackTrace();
         if(clientinfo!=null)
         {
            // togli il cliente dal server
            owner.RemoveClient(clientinfo);
            System.err.println(" with client :\n"+clientinfo.toSimpleString());
            System.err.println("Error informations:\n"+ e);
            System.err.println("Trying to flush data back with a snoop...");
            try {sock.close();}catch(Exception ie){};
         } // se non sappiamo chi  il cliente terminiamo
         else System.err.println(". \nError informations:\n"+ e);

      }
                                               
   }
   
   private boolean InitializeClientConnection()
   {
      boolean closed = false;
      boolean closedsnp = false;
      short start = 0;
      int port = 0;
      long ID = 0;
      Socket answer = null;
      try
      {
         //debug update :
         /* le linee che seguono sono state rimosse
         start = datain.readShort();
         if(start != CacheConstants.CLIENT_BEGINS_SESSION)
         {
            System.err.println("Warning! False step connection.");
            closed = true;
            sock.close();
            return false;
         }
         */
         port = datain.readInt();
         ID = datain.readLong();
         // cliente identificato
         clientinfo = new CacheClientInfo(ID, sock.getInetAddress(),
                                          port, this, sock);
         
         // provvediamo alla risposta di connessione
         // occorre inviare l'ID ricevuto, il nome del server
         // e l'ID del server
         answer = new Socket( sock.getInetAddress(), port );
         DataOutputStream dout = 
                  new DataOutputStream(answer.getOutputStream());
         dout.writeLong(ID);
         dout.writeUTF(owner.getInfo().getName());
         dout.writeLong(owner.getInfo().getID());
         dout.flush();
         closedsnp = true;
         answer.close();
         return true;
      }
      catch( IOException e )
      {   
         System.err.print("An error occoured during initialization");
         if(clientinfo!=null)
         {
            System.err.println(" with client :\n"+clientinfo.toSimpleString());
            System.err.println("Error informations:\n"+ e);
         }
         else System.err.println(". \nError informations:\n"+ e);
         if(!closed) 
         {
            try {sock.close();}
            catch(IOException e2){};
         }
         if(!closedsnp) 
         {
            try {answer.close();}
            catch(IOException e2){};
         }
         return false;
      }
   }   

   private void LineFill(boolean cacheable) throws IOException
   {
      // leggiamo il tag del dato
      ITag tag = readTag(datain);
      //debug
      System.err.println("Linefill of "+
                         clientinfo.getID()+
                         " has "+tag.toString()+ " as argument.");
      // chiediamo di usare il dato come client
      // -> externalsnoop = false
      CacheDataInfo cdi = null;
      try 
      {
         cdi = owner.AskDataForReading(tag, false);
      }  
      catch( CollisionException cce ) {} // non pu avvenire
      catch( ServerPermissionException sce )
      {
         dataout.writeShort(CacheConstants.ACCESS_VIOLATION);
         dataout.flush();
         datain.readShort();
         return;
      }
        
      try
      { 
         //prendiamo le eccezioni solo per rilasciare eventualmente il dato
         // prenotato
      if(cdi==null) // il dato non era nel server ma l'abbiamo
                    // prenotato per uno snoop
      {
         // avviamento dello snoop
         // lista dei server dunque  quella completa
         // mentre quella dei client  vuota
         // proviamo uno snoop sugli altri servers
         cdi = askDataFromLinkedServers(tag, true);// retrieve data
         // se esso  ancora null allora il dato non esiste
         // devo rispondere prontamente al cliente
         if(cdi == null)
         {
            dataout.writeShort(CacheConstants.LF_ANS_MISS);
            dataout.flush();
            owner.ReleaseData(tag);
            datain.readShort();// receive ack
            return; // fine operazione
         }
         else // c' il dato su un altro server
         {
            // memorizziamolo sul nostro server
            // inseriamo il cliente nella lista del dato
            cdi = new CacheDataInfo(cdi.getElem(),
                                    LispUtils.cons( clientinfo, cdi.getClientList() ),
                                    cdi.getServerList(),
                                    cdi.getSource() );
            owner.SaveDataToStorage(cdi);
            // inviamolo al cliente
            boolean share = LispUtils.getListSize(cdi.getClientList()) > 1;
            LineFillInside(cdi.getElem(), cacheable, share);
            owner.ReleaseData(tag);
            return;
         }
      }
      else // il dato  sul server
      {
         // memorizziamolo sul nostro server
         // inseriamo il cliente nella lista del dato
         cdi = new CacheDataInfo(cdi.getElem(),
                                 LispUtils.cons( clientinfo, cdi.getClientList() ),
                                 cdi.getServerList(),
                                 cdi.getSource() );
         owner.SaveDataToStorage(cdi);
         // inviamolo al cliente
         boolean share = LispUtils.getListSize(cdi.getClientList()) > 1;
         LineFillInside(cdi.getElem(), cacheable, share);
         owner.ReleaseData(tag);
         return;
      }
      } //try
      catch( IOException e ) // qualunque cosa accada
      {
         owner.ReleaseData(tag);
         throw e;
      }
   }
   
   private void LineFillInside(ICacheElem elem, boolean cacheable, boolean shareable) throws IOException
   {
      short result = cacheable? 
         (shareable? CacheConstants.LF_ANS_CACHE_SHARE :
                     CacheConstants.LF_ANS_CACHE_EXCL )  :
                     CacheConstants.LF_ANS_UNCACHEABLE ;
      dataout.writeShort(result);
      elem.writeImage(dataout);
      dataout.flush();
      // reading ack
      result = datain.readShort();
      if(result==CacheConstants.LINEFILL_CAUSED_WRITEBACK)
      {
         elem = readCacheElem(datain);
         CacheDataInfo cdi;
         // in che stato  il dato??
         // notare che PER FORZA il dato deve essere 
         // sul server
         // inoltre NON PUO' ESSERE QUELLO APPENA CHIESTO
         // altrimenti NON avremmo avuto una richiesta di linefill
         try {
         cdi = owner.AskDataForWriting(elem.getTag(), true);
         }
         catch( CollisionException c )
         {
            // fatal error
            dataout.writeShort(CacheConstants.NACK);
            dataout.flush();
            throw new IOException() ; //simula un errore
         }
         catch( ServerPermissionException s )
         {
            // fatal error
            dataout.writeShort(CacheConstants.ACCESS_VIOLATION);
            dataout.flush();
            // NON CI SONO ACK DA RACCOGLIERE
            throw new IOException() ; //simula un errore
         }
         // tutto bene
         // il dato del server deve essere aggiornato
         // per eventuali copie sul server uso il nuovo 
         // elemento elem ( se non facessi cos copierei un FakeData!!)
         cdi = new CacheDataInfo(elem, 
                                 cdi.getClientList(),
                                 cdi.getServerList(),
                                 cdi.getSource());
         try
         {
            // aggiorna il server principale anche se si usa la cache
            cdi = setExclusiveUse( cdi, null, clientinfo, true );
         }
         catch( CantSnoopException ce )
         {
            // fatal error
            dataout.writeShort(CacheConstants.NACK);
            dataout.flush();
            throw new IOException() ; //simula un errore
         }
         // i client ed i server sono ora invalidati
         // posso lanciare l'update del server
         owner.SaveDataToStorage(cdi);
         dataout.writeShort(CacheConstants.ACK);
         dataout.flush();
         owner.ReleaseData(elem.getTag());//rilascia il dato writebackato
      }
      // fine operazione
      return;
   }
      
   private void WriteThrough() throws IOException
   {
      // leggiamo il tag del dato
      ITag tag = readTag(datain);
      CacheDataInfo cdi = null;
      boolean newentry = false;// indica se il dato  una new entry
      try
      {
         // chiediamo di usare il dato come client
         // -> externalsnoop = false
         cdi = owner.AskDataForWriting(tag, true);
      }
      catch( CollisionException cce ) 
      {
         // not fatal error
         // non riesco a passare all'uso esclusivo per colpa di una collisione
         
         //debug
         System.err.println("Collision during data seeking.");
         dataout.writeShort(CacheConstants.NACK);
         dataout.flush();
         owner.ReleaseData(tag);
         return;
      } 
      catch( ServerPermissionException sce )
      {
         // not fatal error
         dataout.writeShort(CacheConstants.ACCESS_VIOLATION);
         dataout.flush();
         datain.readShort();
         owner.ReleaseData(tag);
         return;
      }
      try
      {  //prendiamo le eccezioni solo per rilasciare eventualmente il dato
         // prenotato
         // invio l'ACK
         dataout.writeShort(CacheConstants.ACK);
         dataout.flush();
         // leggo il dato
         ICacheElem elem = readCacheElem(datain);
         // il dato era anche sul server ?
         if(cdi!=null)
         {
            // allora uso cdi per settare l'uso esclusivo
            // creo un nuovo elemento di snoop 
            // contenente il nuovo valore
            cdi = new CacheDataInfo( elem, 
                                     cdi.getClientList(),
                                     cdi.getServerList(),
                                     cdi.getSource());
         }
         else // il dato NON era sul server ma lo abbiamo prenotato
         {
            // avviamo una ricerca senza interessarci del dato
            cdi = askDataFromLinkedServers(tag, false); // don't retrieve data
            
            if(cdi != null) // se il dato era su di uno dei servers
            {
               // allora sar necessario invalidare ogni server 
               // della lista ritornata, sorgente compresa
               // tanto se siamo il master essa  a null
               cdi = new CacheDataInfo(elem, 
                                       LispUtils.EmptyList(),
                                       cdi.getServerList(),
                                       cdi.getSource());
            }
            else // se il dato  una new entry
            {
               // in tal caso non si deve aggiornare nessuno, solo il master
               // se noi siamo uno slave
               CacheServerInfo src = null;
               if(!owner.getMaster())
                  src = (CacheServerInfo)(owner.getBros().car());
               cdi = new CacheDataInfo( elem,
                                        LispUtils.EmptyList(),
                                        LispUtils.EmptyList(),
                                        src);
            }
         }
         CacheDataInfo ncdi = null;
         try
         {
            // lancia uno snoop 
            // invalidante con aggiornamento
            ncdi = setExclusiveUse(cdi, null, clientinfo, true);
         }
         catch( CantSnoopException cnt )
         {
            // not fatal error
            // non riesco a passare all'uso esclusivo
            //debug
            System.err.println("Collision during setExclusiveUse.");
            dataout.writeShort(CacheConstants.NACK);
            dataout.flush();
            owner.ReleaseData(tag);
            return;
         }
         // salvo il dato
         owner.SaveDataToStorage(ncdi);
         // invio l'ack
         dataout.writeShort(CacheConstants.ACK);
         dataout.flush();
         owner.ReleaseData(tag);
         return;
      }
      catch( IOException e ) // qualunque cosa accada
      {
         owner.ReleaseData(tag);
         throw e;
      }
   }
   
   private void WriteExclusive() throws IOException
   {
      // leggiamo il tag del dato
      ITag tag = readTag(datain);
      CacheDataInfo cdi = null;
      try
      {
         // chiediamo di usare il dato come client
         // -> externalsnoop = false
         cdi = owner.AskDataForWriting(tag, true);
      }
      catch( CollisionException cce ) 
      {
         // not fatal error
         // non riesco a passare all'uso esclusivo
         dataout.writeShort(CacheConstants.NACK);
         dataout.flush();
         owner.ReleaseData(tag);
         return;
      } 
      catch( ServerPermissionException sce )
      {
         // not fatal error - cache is still valid
         dataout.writeShort(CacheConstants.ACCESS_VIOLATION);
         dataout.flush();
         datain.readShort();
         owner.ReleaseData(tag);
         return;
      }
      try
      {  //prendiamo le eccezioni solo per rilasciare eventualmente il dato
         // prenotato
         
         // il dato DEVE essere sul server
         if(cdi==null)
         {
            // in caso contrario vietiamo il WE
            // not fatal error - cache is still valid
            dataout.writeShort(CacheConstants.ACCESS_VIOLATION);
            dataout.flush();
            datain.readShort();
            owner.ReleaseData(tag);
            return;
         }
         // chiediamo l'uso esclusivo del dato
         try
         {
            // il master non aggiorna gli slaves
            // lo slave invia al master una copia del dato vecchio
            cdi = setExclusiveUse(cdi, null, clientinfo, false);
         }
         catch( CantSnoopException cnt )
         {
            // not fatal error
            // non riesco a passare all'uso esclusivo
            dataout.writeShort(CacheConstants.NACK);
            dataout.flush();
            owner.ReleaseData(tag);
            return;
         }
         // non salvo il dato ma lo invalido con nuova sorgente
         // la cache cliente
         owner.MarkDataAsInvalidFromCache(cdi.getElem().getTag(), clientinfo);
         // invio l'ack
         dataout.writeShort(CacheConstants.ACK);
         dataout.flush();
         owner.ReleaseData(tag);
         return;
      }
      catch( IOException e ) // qualunque cosa accada
      {
         owner.ReleaseData(tag);
         throw e;
      }
   }
   
   private void Update() throws IOException
   {
      IUpdateElem upd = readUpdateElem(datain);
      ITag tag = upd.getTag();
      CacheDataInfo cdi = null;
      boolean datafromother;
      try
      {
         // chiediamo di usare il dato come client
         // -> externalsnoop = false
         cdi = owner.AskDataForWriting(tag, true);
      }
      catch( CollisionException cce ) 
      {
         // not fatal error
         // non riesco a passare all'uso esclusivo
         dataout.writeShort(CacheConstants.NACK);
         dataout.flush();
         owner.ReleaseData(tag);
         return;
      } 
      catch( ServerPermissionException sce )
      {
         // not fatal error
         dataout.writeShort(CacheConstants.ACCESS_VIOLATION);
         dataout.flush();
         datain.readShort();
         owner.ReleaseData(tag);
         return;
      }
      try
      {  //prendiamo le eccezioni solo per rilasciare eventualmente il dato
         // prenotato
         datafromother = cdi == null;
         if(datafromother) cdi = askDataFromLinkedServers(tag, true);
         if(cdi == null) // still null : niente update
         {
            dataout.writeShort(CacheConstants.NACK);
            dataout.flush();
            owner.ReleaseData(tag);
            return; // fine operazione
         }
         try
         {
            //prova ad aggiornare il dato
            ICacheElem elem = updateIsCorrect(cdi.getElem(),upd);
            // se il dato nuovo  corretto allora invalida gli altri server
            CacheDataInfo ncdi = new CacheDataInfo(elem, 
                                                   cdi.getClientList(),
                                                   cdi.getServerList(),
                                                   cdi.getSource());
            // chiediamo l'uso esclusivo del dato
            try
            {
               // nota : AGGIORNIAMO IL DATO
               ncdi = setExclusiveUse(ncdi, null, clientinfo, true);
            }
            catch( CantSnoopException cnt )
            {
               // not fatal error
               // non riesco a passare all'uso esclusivo
               dataout.writeShort(CacheConstants.NACK);
               dataout.flush();
               owner.ReleaseData(tag);
               return;
            }
            // salvo il dato
            owner.SaveDataToStorage(ncdi);
            // invio l'ack
            dataout.writeShort(CacheConstants.ACK);
            dataout.flush();
            owner.ReleaseData(tag);
            return;
         }
         catch(InconsistentUpdateException iue)
         {
            // not fatal error
            dataout.writeShort(CacheConstants.CANT_UPDATE_INCONSISTENCE);
            dataout.flush();
            datain.readShort();
            owner.ReleaseData(tag);
            return;
         }
      }
      catch( IOException e ) // qualunque cosa accada
      {
         owner.ReleaseData(tag);
         throw e;
      }
   }
   
   // disconnessione server-side
   private void LaunchFinalSnoopFlush() throws IOException
   {
      Socket snoopsock = null;
      boolean open = false;
      try{
      // dobbiamo lanciare uno snoop sul cliente
      // creiamo una socket di snoop
      snoopsock = new Socket(clientinfo.getAddress(),
                                    clientinfo.getSnoopPort());
      open = true;
      // settaggio opzioni socket
      snoopsock.setTcpNoDelay(CacheConstants.TCPNODELAY);
      snoopsock.setSoTimeout(CacheConstants.TIMEOUT);
      snoopsock.setSoLinger(CacheConstants.LINGER, CacheConstants.LINGERTIME);
      // imposto gli stream
      DataInputStream din = new DataInputStream( snoopsock.getInputStream() );
      DataOutputStream dout = new DataOutputStream( snoopsock.getOutputStream() );
      // lanciamo il comando
      dout.writeLong(owner.getInfo().getID());
      dout.writeShort(CacheConstants.SERVER_IS_CLOSING);
      dout.flush();
      // leggiamo la risposta
      short answer = din.readShort();
      ICacheElem[] elems = null;
      int bodycount = 0;
      if(answer == CacheConstants.HITM)
      {
         // leggi quanti
         bodycount = din.readInt();
         elems = new ICacheElem[bodycount];
         for(int i = 0; i< bodycount ; i++)
            elems[i] = readCacheElem(din);
      }
      if(answer == CacheConstants.MISS)
         return; /// in caso di miss terminiamo qui
      // questi dati sono anche sul server
      // dobbiamo aggiornare eventuali altri server
      // i client no poich ci pensano gi gli atri CacheServingThreads
      for(int i = 0; i<bodycount; i++)
      {
         CacheDataInfo dinfo;
         try
         {
            dinfo = owner.AskDataForWriting(elems[i].getTag(),false);
            dinfo = new CacheDataInfo(elems[i],//nuovo dato
                                      LispUtils.EmptyList(),//nessun cliente
                                      dinfo.getServerList(),
                                      dinfo.getSource() );
            // aggiorniamo gli altri
            dinfo = setExclusiveUse(dinfo,null,clientinfo,true);
         }
         catch(CacheException ce)
         {
            // pazienza
            dout.writeShort(CacheConstants.NACK);
            dout.flush();
            snoopsock.close();
            open = false;
         }
      }
      dout.writeShort(CacheConstants.ACK);
      dout.flush();
      snoopsock.close();
      open = false;
      // nota : a questo punto sarebbe meglio 
      // far fallire i comandi client 
      // che sono eventualmente rimasti pendenti
      // con una violazione di accesso
      boolean ct = true;
      while(ct)
      {
         try
         {
            short c =datain.readShort(); // legge il comando
            sock.setSoTimeout(CacheConstants.TIMEOUT);
            short rt = (c==CacheConstants.CLIENT_DISCONNECTING)?
                        CacheConstants.ACK : CacheConstants.ACCESS_VIOLATION ;
            dataout.writeShort(rt);
            dataout.flush();
            if(c!=CacheConstants.CLIENT_DISCONNECTING) datain.readShort();
            sock.close();
         }
         catch(InterruptedIOException ier){ ct = false; sock.close(); }
      }
      
      }
      catch(IOException e)
      { if(open) snoopsock.close(); sock.close(); throw e ; }
   
   }
   
   // disconnessione client-side
   private void Disconnecting() throws IOException
   {
      // il flush  gi avvenuto
      // limitiamoci all'ACK
      dataout.writeShort(CacheConstants.ACK);
      dataout.flush();
      sock.close(); // chiusura della connessione
   }
         
} // CacheServingThread
