/** File MagController.java
* Classe Java che implementa 
* i metodi per garantire la consistenza
* dei dati della cache e i metodi che 
* leggono i dati dalla rete.
* (CLMDNR = Consistency
*           Logic
*           Manager
*           Data
*           Net
*           Receiver 
@author Claudio Forlivesi
@version 1.0, 15/2/2000 
*/

import java.io.*;

public class MagController extends CacheCLMDNR 
{
   public ICacheElem updateIsCorrect( ICacheElem oldelem, IUpdateElem news ) 
                                 throws InconsistentUpdateException
   {
      StoredMat mat = null;
      UpdateOp op = null;
      
      if(oldelem instanceof StoredMat)
         mat = (StoredMat)oldelem;
      else 
         throw new InconsistentUpdateException("Errore nel "+
                   "tentativo di aggiornare un dato che non  " +
                   "una voce di magazzino:\n" +
                   oldelem.toString() );

      int newquantity = mat.getQuantity();
      int newcost = mat.getCost();
      
      if(news instanceof UpdateOp)
         op = (UpdateOp)news;
      else 
         throw new InconsistentUpdateException("Errore nel "+
                   "tentativo di aggiornare la voce di magazzino\n " +
                   oldelem.toString() +
                   "\ncon una operazione non consentita per tale voce:\n" +
                   news.toString() );
      
      if((op.getOperation() & MagConstants.COSTOP)!=0)
      {
         int updcost = op.getOpCost();
         switch(op.getOperation())
         {
         case MagConstants.ADDCOST:
            newcost += updcost ; break;
         case MagConstants.SUBCOST:
            if(newcost < updcost)
               throw new InconsistentUpdateException("Errore nel " +
               "tentativo di ridurre il prezzo di "+ updcost +
               "$ del materiale "+ mat.getTag() +" che costa " +
               mat.getCost() +".\n");
            newcost -= updcost ; break;
         case MagConstants.NEWCOST:
            if(updcost < 0)
               throw new InconsistentUpdateException("Errore nel " +
               "tentativo di applicare un prezzo di "+ updcost +
               "$ al materiale "+ mat.getTag() +".\n");
            newcost = updcost ; break;
         }
      }
      
      if((op.getOperation() & MagConstants.QUANTITYOP)!=0)
      {
         int updquant = op.getOpQuantity();
         switch(op.getOperation())
         {
         case MagConstants.ADDQUANTITY:
            newquantity += updquant ; break;
         case MagConstants.SUBQUANTITY:
            if(newquantity < updquant)
               throw new InconsistentUpdateException("Errore nel " +
               "tentativo di diminuire le scorte di "+ updquant +
               " unit del materiale "+ mat.getTag() +
               " di cui si hanno "+ newquantity +
               " unit immagazzinate.\n");
            newquantity -= updquant ; break;
         case MagConstants.NEWQUANTITY:
            if(updquant < 0)
               throw new InconsistentUpdateException("Errore nel " +
               "tentativo di immagazzinare "+ updquant +
               " unit di materiale "+ mat.getTag() +".\n");
            newquantity = updquant ; break;
         }
      }
         
      if(((op.getOperation() & MagConstants.QUANTITYOP)==0) &&
         ((op.getOperation() & MagConstants.COSTOP)==0))
            throw new InconsistentUpdateException("Si  tentato di "+
            "applicare alla voce di magazzino " + mat.getTag() +
            " una operazione sconosciuta : "+
            Integer.toString(op.getOperation(), 16).toUpperCase()+
            ".\n"  );
      
      return new StoredMat(mat.getCode(),
                           newquantity, newcost);
                           
   }
                                    
   public ITag readTag( DataInputStream datain ) 
                                 throws IOException
   {
      short type = datain.readShort();
      switch(type)
      {
      case MagConstants.MAGTAG:
         char[] c = new char[MagConstants.MAT_CODE_LEN];
         for (int i = 0 ; i < MagConstants.MAT_CODE_LEN ; i++)
            c[i] = datain.readChar();
         return new MagTag(c);

      case MagConstants.DESCRTAG:
         return new DescrTag(datain.readUTF());
         
      default:
         throw new UnknownTypeException("Tag", type);
      }
   }
                                    
   public IUpdateElem readUpdateElem( DataInputStream datain ) 
                                 throws IOException
   {
      short operation = datain.readShort();
      int opcost = 0;
      int opquantity = 0;
      
      if((operation & MagConstants.QUANTITYOP)!=0)
         opquantity = datain.readInt();

      if((operation & MagConstants.COSTOP)!=0)
         opcost = datain.readInt();

      if(((operation & MagConstants.QUANTITYOP)==0) &&
         ((operation & MagConstants.COSTOP)==0))
            throw new UnknownTypeException("update operation",
                                           operation );
                                           
      char[] c = new char[MagConstants.MAT_CODE_LEN];
      for (int i = 0 ; i < MagConstants.MAT_CODE_LEN ; i++)
         c[i] = datain.readChar();
      
      return new UpdateOp(new MagTag(c),
                          operation,
                          opquantity,
                          opcost);
   }
                                    
   public ICacheElem readCacheElem( DataInputStream datain ) 
                                 throws IOException
   {
      short type = datain.readShort();
      switch(type)
      {
      case MagConstants.MATDESCR:
         String matname = datain.readUTF();
         char[] c = new char[MagConstants.MAT_CODE_LEN];
         for (int i = 0 ; i < MagConstants.MAT_CODE_LEN ; i++)
            c[i] = datain.readChar();
         String matdescr = datain.readUTF();
         return new MatDescr(matname, matdescr, new MagTag(c) );
         
      case MagConstants.STOREDMAT:
         int quantity = datain.readInt();
         int cost = datain.readInt();
         char[] cd = new char[MagConstants.MAT_CODE_LEN];
         for (int i = 0 ; i < MagConstants.MAT_CODE_LEN ; i++)
            cd[i] = datain.readChar();
         return new StoredMat(new MagTag(cd), quantity, cost);
         
      default:
         throw new UnknownTypeException("Data", type);
      }
   }                                 

} // MagController  
 


